/*
  2.4 to 2.7,2.8 converter
*/
#include <stdio.h>
#include <string.h>
#include <utmp.h>

#define STR_LEN       256
#define STR_EMAIL_LEN 36
#define STR_USER_LEN  (UT_NAMESIZE + 1)


/** 2.4 ************************************************/
typedef struct
{
  char         version[5];
  int          run_level;
  unsigned int sleep_time;
  int          no_of_warnings;
  int          midnreset;
  int          clobber_sig;
} tusettings_hr_24;

typedef struct
{
  char         user[STR_USER_LEN];
  uid_t        uid;
  pid_t        pid;
  char         active;
  time_t       total_time_used;
  time_t       daily_time_used;
  int          no_ttime_warnings;
  int          no_dtime_warnings;
  int          old_jiffies;
  int          jiffies;
  unsigned int rss;
  int          no_cpu_warnings;
  int          no_mem_warnings;
} tutmp_24;

/** 2.8 ************************************************/
typedef struct
{
  uid_t  uid;
  char   email[STR_EMAIL_LEN];
  time_t total_time_limit;
  time_t daily_time_limit;
  int    nice_factor;
  int    cpu_limit;
  int    mem_limit;
  int    exempt;
} tusettings_28;

typedef struct
{
  int           run_level;
  unsigned int  sleep_time;
  int           no_of_warnings;
  short         clobber_sig;
  short         min_free_dialins;
  int           def_usettings_enab;
  tusettings_28 def_usettings;
} tusettings_hr_28;

typedef struct
{
  int yesterday;
} tutmp_hr_28;

typedef struct
{
  char         user[STR_USER_LEN];
  uid_t        uid;
  pid_t        pid;
  char         active;
  time_t       total_time_used;
  time_t       daily_time_used;
  int          no_ttime_warnings;
  int          no_dtime_warnings;
  int          old_jiffies;
  int          jiffies;
  unsigned int rss;
  int          no_cpu_warnings;
  int          no_mem_warnings;
  float        nice_factor;
} tutmp_28;

int main(void)
{
  FILE *fp_r,*fp_w;
  tusettings_hr_24 usettings_hr_24;
  tusettings_hr_28 usettings_hr_28;
  tusettings_28    usettings_28;
  tutmp_hr_28      utmp_hr_28;
  tutmp_24         utmp_24;
  tutmp_28         utmp_28;

  if((fp_r = fopen("clobberd.cf","rb")) != NULL)
  {
    if((fp_w = fopen("clobberd.cf.2.8","wb")) != NULL)
    {
      fread(&usettings_hr_24,sizeof(tusettings_hr_24),1,fp_r);

      usettings_hr_28.run_level =          usettings_hr_24.run_level;
      usettings_hr_28.sleep_time =         usettings_hr_24.sleep_time;
      usettings_hr_28.no_of_warnings =     usettings_hr_24.no_of_warnings;
      usettings_hr_28.clobber_sig =        usettings_hr_24.clobber_sig;
      usettings_hr_28.min_free_dialins =   0;
      usettings_hr_28.def_usettings_enab = 0;

      usettings_hr_28.def_usettings.uid =              65535;
      strcpy(usettings_hr_28.def_usettings.email,      "");
      usettings_hr_28.def_usettings.total_time_limit = 0;
      usettings_hr_28.def_usettings.daily_time_limit = 0;
      usettings_hr_28.def_usettings.nice_factor =      0;
      usettings_hr_28.def_usettings.cpu_limit =        0;
      usettings_hr_28.def_usettings.mem_limit =        0;
      usettings_hr_28.def_usettings.exempt =           0;

      fwrite(&usettings_hr_28,sizeof(tusettings_hr_28),1,fp_w);

      while(fread(&usettings_28,sizeof(tusettings_28),1,fp_r) == 1)
        fwrite(&usettings_28,sizeof(tusettings_28),1,fp_w);

      fclose(fp_w);
    }
    fclose(fp_r);
  }


  if((fp_r = fopen("clobberd.db","rb")) != NULL)
  {
    if((fp_w = fopen("clobberd.db.2.8","wb")) != NULL)
    {
      utmp_hr_28.yesterday = 0;

      fwrite(&utmp_hr_28,sizeof(tutmp_hr_28),1,fp_w);

      while(fread(&utmp_24,sizeof(tutmp_24),1,fp_r) == 1)
      {
        strcpy(utmp_28.user,        utmp_24.user);
        utmp_28.uid =               utmp_24.uid;
        utmp_28.pid =               utmp_24.pid;
        utmp_28.active =            utmp_24.active;
        utmp_28.total_time_used =   utmp_24.total_time_used;
        utmp_28.daily_time_used =   utmp_24.daily_time_used;
        utmp_28.no_ttime_warnings = utmp_24.no_ttime_warnings;
        utmp_28.no_dtime_warnings = utmp_24.no_dtime_warnings;
        utmp_28.old_jiffies =       utmp_24.old_jiffies;
        utmp_28.jiffies =           utmp_24.jiffies;
        utmp_28.rss =               utmp_24.rss;
        utmp_28.no_cpu_warnings =   utmp_24.no_cpu_warnings;
        utmp_28.no_mem_warnings =   utmp_24.no_mem_warnings;
        utmp_28.nice_factor =       0;

        fwrite(&utmp_28,sizeof(tutmp_28),1,fp_w);
      }

      fclose(fp_w);
    }
    fclose(fp_r);
  }

  return 1;
}
