/***************************************************************************
****************************************************************************
****************************************************************************
*
* Pwed - By Jason Nunn
* FREEWARE. Authorship Reserved 1996
*
* Using Xforms
*
* Snail: 32 Rothdale Road, Moil, Darwin, NT, 0810, Australia
*
****************************************************************************
****************************************************************************
***************************************************************************/
#include <stdio.h>
#include "forms.h"
#include <stdlib.h>
#include "pwedif.h"
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#define VERSION  "0.0"
#define DATED    __DATE__
#define WELCOME \
"Welcome to Pwed " VERSION ", Compiled: " DATED \
"\nby Jason Nunn. 1997\n" \
"Contact: 32 Rothdale Rd, Moil Darwin NT 0810, AUSTRALIA\n" \
"         jsno@stormfront.com.au,jsno@dayworld.net.au\n" \
"         http://stormfront.com.au/~jsno/\n\n"

#define UID_USER_MIN 501
#define UID_USER_MAX 65535

#define SHORT_STR_LEN 20
#define LONG_STR_LEN  256

typedef struct _tgr
{
  char  name[SHORT_STR_LEN];
  gid_t gid;
} tgr;

typedef struct _tgroup_ll
{
  tgr               *curr;
  struct _tgroup_ll *next;
} tgroup_ll;

typedef struct _tpw
{
  char  name[SHORT_STR_LEN];
  char  passwd[SHORT_STR_LEN];
  uid_t uid;
  gid_t gid;
  char  gecos[LONG_STR_LEN];
  char  dir[LONG_STR_LEN];
  char  shell[LONG_STR_LEN];
} tpw;

typedef struct _tpasswd_ll
{
  tpw                *curr;
  struct _tpasswd_ll *next;
} tpasswd_ll;

tgroup_ll group_ll_root = {NULL,NULL};
tpasswd_ll passwd_ll_root = {NULL,NULL};

FD_pwedif *fd_pwedif;

/***************************************************************************
*
***************************************************************************/
void load_gr(void)
{
  int f = 1;

  void load_entry(tgroup_ll *entry)
  {
    struct group *gr;

    if((gr = getgrent()) != NULL)
    {
      fl_addto_choice(fd_pwedif->gid_choice,gr->gr_name);
      if((entry->curr = malloc(sizeof(tgr))) != NULL)
      {
        strncpy(entry->curr->name,gr->gr_name,SHORT_STR_LEN);
        entry->curr->gid = gr->gr_gid;
        if((entry->next = malloc(sizeof(tpasswd_ll))) != NULL)
        {
          entry->next->curr = NULL;
          entry->next->next = NULL;
          load_entry(entry->next);
        }
        else
          f = 0;
      }
      else
        f = 0;
    }
  }
  setgrent();
  load_entry(&group_ll_root);
  endgrent();
  if(!f) fl_show_alert("load_gr()","Can't Malloc()","any more memory", 0);
}

int get_gr_choice_index(gid_t gid)
{
  int r = 0,pos = 0;
  
  void find_entry(tgroup_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(entry->curr->gid == gid)
      {
        r = pos + 1;
        return;
      }
    }
    if(entry->next != NULL)
    {
      pos++;
      find_entry(entry->next);
    }
  }
  find_entry(&group_ll_root);
  return r;
}

/***************************************************************************
*
***************************************************************************/
int insert_pw_entry(struct passwd *pw)
{
  int f = 1;

  void search_insert_entry(tpasswd_ll *entry)
  {
    if(entry->curr == NULL)
    {
      if((entry->curr = malloc(sizeof(tpw))) != NULL)
      {
        strncpy(entry->curr->name,pw->pw_name,SHORT_STR_LEN);
        strncpy(entry->curr->passwd,pw->pw_passwd,SHORT_STR_LEN);
        entry->curr->uid = pw->pw_uid;
        entry->curr->gid = pw->pw_gid;
        strncpy(entry->curr->gecos,pw->pw_gecos,LONG_STR_LEN);
        strncpy(entry->curr->dir,pw->pw_dir,LONG_STR_LEN);
        strncpy(entry->curr->shell,pw->pw_shell,LONG_STR_LEN);
      }
      else
        f = 0;
    }
    else
    {
      if(entry->next != NULL)
      {
        search_insert_entry(entry->next);
      }
      else
        if((entry->next = malloc(sizeof(tpasswd_ll))) != NULL)
        {
          entry->next->curr = NULL;
          entry->next->next = NULL;
          search_insert_entry(entry->next);
        }
        else
          f = 0;
    }
  }
  search_insert_entry(&passwd_ll_root);
  if(!f) fl_show_alert("insert_pw_entry()","Can't Malloc()","any more memory",0);
  return f;
}

void delete_pw_entry(char *str)
{
  int f = 0;

  void search_delete_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(!strcmp(entry->curr->name,str))
      {
        free(entry->curr);
        entry->curr = NULL;
        f = 1;
      }
    }
    if((!f) && (entry->next != NULL)) search_delete_entry(entry->next);
  }
  search_delete_entry(&passwd_ll_root);
}

void update_pw_entry(struct passwd *pw,char *str)
{
  int f = 0;

  void search_update_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(!strcmp(entry->curr->name,str))
      {
        strncpy(entry->curr->name,pw->pw_name,SHORT_STR_LEN);
        strncpy(entry->curr->passwd,pw->pw_passwd,SHORT_STR_LEN);
        entry->curr->gid = pw->pw_gid;
        strncpy(entry->curr->gecos,pw->pw_gecos,LONG_STR_LEN);
        strncpy(entry->curr->dir,pw->pw_dir,LONG_STR_LEN);
        strncpy(entry->curr->shell,pw->pw_shell,LONG_STR_LEN);
        f = 1;
      }
    }
    if((!f) && (entry->next != NULL)) search_update_entry(entry->next);
  }
  search_update_entry(&passwd_ll_root);
}

void update_shell_pw_entry(char *shell,gid_t gid)
{
  void update_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
      if(entry->curr->gid == gid)
        strncpy(entry->curr->shell,shell,LONG_STR_LEN);
    if(entry->next != NULL) update_entry(entry->next);
  }
  update_entry(&passwd_ll_root);
}

void load_pw_table(void)
{
  struct passwd *pw;

  setpwent();
  while((pw = getpwent()) != NULL)
    if(!insert_pw_entry(pw)) break;
  endpwent();
}

void load_browser(void)
{
  void load_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
      fl_addto_browser(fd_pwedif->passwd_browser,entry->curr->name);
    if(entry->next != NULL) load_entry(entry->next);
  }
  load_entry(&passwd_ll_root);
  fl_set_browser_topline(fd_pwedif->passwd_browser,1);
}

tpw *find_pw_entry(char *user)
{
  tpw *r = NULL;

  void find_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(!strcmp(entry->curr->name,user))
      {
        r = entry->curr;
        return;
      }
    }
    if(entry->next != NULL) find_entry(entry->next);
  }
  find_entry(&passwd_ll_root);
  return r;
}

uid_t get_unique_uid(void)
{
  uid_t uid;
  int f;

  void find_uid(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
    {
      if(entry->curr->uid == uid)
      {
        f = 1;
        return;
      }
    }
    if(entry->next != NULL) find_uid(entry->next);
  }
  for(uid = UID_USER_MIN;uid < UID_USER_MAX;uid++)
  {
    f = 0;
    find_uid(&passwd_ll_root);
    if(!f) break;
  }
  return uid;
}

/***************************************************************************
*
***************************************************************************/
void passwd_browser_cb(FL_OBJECT * ob,long q)
{
  register int r;
  char *user;
  tpw *entry;

  r = fl_get_browser(ob);
  if(r <= 0) return;
  user = (char *)fl_get_browser_line(ob,r);
  if((entry = find_pw_entry(user)) != NULL)
  {
    char str[10];

    fl_set_input(fd_pwedif->user_input,entry->name);
    fl_set_input(fd_pwedif->passwd_input,entry->passwd);
    sprintf(str,"%d",entry->uid);
    fl_set_input(fd_pwedif->uid_input,str);
    fl_set_choice(fd_pwedif->gid_choice,get_gr_choice_index(entry->gid));
    fl_set_input(fd_pwedif->name_input,entry->gecos);
    fl_set_input(fd_pwedif->dir_input,entry->dir);
    fl_set_input(fd_pwedif->shell_input,entry->shell);
  }
}

/***************************************************************************
*
***************************************************************************/
void get_inputs(struct passwd *pw)
{
  char *str;
  
  pw->pw_name = (char *)fl_get_input(fd_pwedif->user_input);
  pw->pw_passwd = (char *)fl_get_input(fd_pwedif->passwd_input);
  str = (char *)fl_get_choice_text(fd_pwedif->gid_choice);
  if(str != NULL)
  {
    struct group *gr;
      
    if((gr = getgrnam(str)) != NULL) pw->pw_gid = gr->gr_gid;
  }
  pw->pw_gecos = (char *)fl_get_input(fd_pwedif->name_input);
  pw->pw_dir = (char *)fl_get_input(fd_pwedif->dir_input);
  pw->pw_shell = (char *)fl_get_input(fd_pwedif->shell_input);
}

void passwd_add_cb(FL_OBJECT * ob,long q)
{
  struct passwd pw;

  if((pw.pw_uid = get_unique_uid()) != UID_USER_MAX)
  {
    get_inputs(&pw);
    insert_pw_entry(&pw);
    fl_addto_browser(fd_pwedif->passwd_browser,pw.pw_name);
  }
}

void passwd_delete_cb(FL_OBJECT * ob,long q)
{ 
  register int r;  

  r = fl_get_browser(fd_pwedif->passwd_browser);
  if(r <= 0) return;
  delete_pw_entry((char *)fl_get_browser_line(fd_pwedif->passwd_browser,r));
  fl_delete_browser_line(fd_pwedif->passwd_browser,r);

  fl_set_input(fd_pwedif->user_input,"");
  fl_set_input(fd_pwedif->passwd_input,"");
  fl_set_input(fd_pwedif->uid_input,"");
  fl_set_choice(fd_pwedif->gid_choice,0);
  fl_set_input(fd_pwedif->name_input,"");
  fl_set_input(fd_pwedif->dir_input,"");
  fl_set_input(fd_pwedif->shell_input,"");
}

void passwd_update_cb(FL_OBJECT * ob,long q)
{
  register int r;
  struct passwd pw;
  char *str;

  get_inputs(&pw);
  r = fl_get_browser(fd_pwedif->passwd_browser);
  if(r <= 0) return;
  str = (char *)fl_get_browser_line(fd_pwedif->passwd_browser,r);
  update_pw_entry(&pw,str);
  fl_replace_browser_line(fd_pwedif->passwd_browser,r,pw.pw_name);
}

void passwd_update_shell_cb(FL_OBJECT * ob,long q)
{
  register int r;
  char *str;

  r = fl_get_browser(fd_pwedif->passwd_browser);
  if(r <= 0) return;

  str = (char *)fl_get_choice_text(fd_pwedif->gid_choice);
  if(str != NULL)
  {
    struct group *gr;
  
    if((gr = getgrnam(str)) != NULL)
      update_shell_pw_entry((char *)fl_get_input(fd_pwedif->shell_input),gr->gr_gid);
  }
}

void abort_cb(FL_OBJECT * ob,long q)
{
  fl_hide_form(fd_pwedif->pwedif);
  printf("Aborted.. No changes done.\n");
  exit(0);
}

void save_cb(FL_OBJECT * ob,long q)
{
  FILE *fp;
  struct passwd pw;

  void write_entry(tpasswd_ll *entry)
  {
    if(entry->curr != NULL)
    {
      pw.pw_name = entry->curr->name;
      pw.pw_passwd = entry->curr->passwd;
      pw.pw_uid = entry->curr->uid;
      pw.pw_gid = entry->curr->gid;
      pw.pw_gecos = entry->curr->gecos;
      pw.pw_dir = entry->curr->dir;
      pw.pw_shell = entry->curr->shell;
      putpwent(&pw,fp);
    }
    if(entry->next != NULL) write_entry(entry->next);
  }

  if(rename("/etc/passwd","/etc/passwd.old") != -1)
  {
    if((fp = fopen("/etc/passwd","w")) != NULL)
      write_entry(&passwd_ll_root);
    else
      fl_show_alert("save_cb()","Can't fopen()","", 0);
    fl_hide_form(fd_pwedif->pwedif);
    printf("Updated /etc/passwd..\n");
    exit(0);
  }
  else
    fl_show_alert("save_cb()","Can't rename()","passwd to passwd.old", 0);
}

/***************************************************************************
*
***************************************************************************/
FD_pwedif *create_form_pwedif(void)
{
  FL_OBJECT *obj;
  FD_pwedif *fdui = (FD_pwedif *) fl_calloc(1, sizeof(*fdui));

  fdui->pwedif = fl_bgn_form(FL_NO_BOX, 440, 340);
  obj = fl_add_box(FL_UP_BOX,0,0,440,340,"");
  fdui->passwd_browser = obj = fl_add_browser(FL_HOLD_BROWSER,10,10,130,320,"");
    fl_set_object_boxtype(obj,FL_UP_BOX);
    fl_set_object_color(obj,FL_COL1,FL_WHEAT);
    fl_set_browser_fontstyle(obj,FL_FIXED_STYLE);
    fl_set_browser_fontsize(obj,16);
    fl_set_object_callback(obj,passwd_browser_cb,0);
  fdui->user_input = obj = fl_add_input(FL_NORMAL_INPUT,190,20,120,25,"User");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->passwd_input = obj = fl_add_input(FL_NORMAL_INPUT,190,50,120,25,"Pass");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->uid_input = obj = fl_add_input(FL_INT_INPUT,190,80,70,25,"UID");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->gid_choice = obj = fl_add_choice(FL_NORMAL_CHOICE,190,110,120,25,"GID");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
    fl_set_choice_fontstyle(obj,FL_FIXED_STYLE);
    fl_set_choice_fontsize(obj,16);
  fdui->name_input = obj = fl_add_input(FL_NORMAL_INPUT,190,140,240,25,"Name");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->dir_input = obj = fl_add_input(FL_NORMAL_INPUT,190,170,240,25,"Home");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->shell_input = obj = fl_add_input(FL_NORMAL_INPUT,190,200,240,25,"Shell");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,14);
    fl_set_object_lstyle(obj,FL_FIXED_STYLE+FL_SHADOW_STYLE);
  fdui->save_button = obj = fl_add_button(FL_NORMAL_BUTTON,250,280,170,50,"Save");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
    fl_set_object_callback(obj,save_cb,0);
  fdui->add_button = obj = fl_add_button(FL_NORMAL_BUTTON,150,230,60,40,"Add");
    fl_set_object_lcol(obj,FL_YELLOW);
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
    fl_set_object_callback(obj,passwd_add_cb,0);
  fdui->delete_button = obj = fl_add_button(FL_NORMAL_BUTTON,220,230,60,40,"Delete");
    fl_set_object_lcol(obj,FL_YELLOW);
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
    fl_set_object_callback(obj,passwd_delete_cb,0);
  fdui->update_button = obj = fl_add_button(FL_NORMAL_BUTTON,290,230,60,40,"Update");
    fl_set_object_lcol(obj,FL_YELLOW);
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
    fl_set_object_callback(obj,passwd_update_cb,0);
  fdui->abort_button = obj = fl_add_button(FL_NORMAL_BUTTON,150,280,90,50,"Abort");
    fl_set_object_lcol(obj,FL_ORCHID);
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
    fl_set_object_callback(obj,abort_cb,0);
  obj = fl_add_box(FL_DOWN_BOX,320,20,110,110,"");
  obj = fl_add_text(FL_NORMAL_TEXT,331,23,90,35,"PWED!!");
    fl_set_object_lcol(obj,FL_CHARTREUSE);
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,325,77,100,20,"By Jason Nunn");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,324,99,99,19,"1997, for Xforms");
    fl_set_object_lcol(obj,FL_WHITE);
    fl_set_object_lsize(obj,FL_NORMAL_SIZE);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,336,54,80,20,"Passwd editor");
    fl_set_object_lcol(obj,FL_WHEAT);
    fl_set_object_lsize(obj,11);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_SHADOW_STYLE);
  fdui->update_shell_button = obj = fl_add_button(FL_NORMAL_BUTTON,360,230,60,40,"U Shell");
    fl_set_object_lcol(obj,FL_YELLOW);
    fl_set_object_lsize(obj,FL_MEDIUM_SIZE);
    fl_set_object_callback(obj,passwd_update_shell_cb,0);
    fl_set_object_lstyle(obj,FL_NORMAL_STYLE+FL_ENGRAVED_STYLE);
  fl_end_form();

  return fdui;
}

/***************************************************************************
*
***************************************************************************/
int main(int argc, char *argv[])
{
  printf(WELCOME);
  fl_initialize(&argc,argv,0,0,0);
  fd_pwedif = create_form_pwedif();
  load_gr();
  load_pw_table();
  load_browser();
  fl_show_form(fd_pwedif->pwedif,FL_PLACE_MOUSE,FL_FULLBORDER,
    "Pwed - Passwd Editor " VERSION);
  while(1) fl_do_forms();
  return 1;
}
